import logging

from spaceone.core import config
from spaceone.core.error import *

_LOGGER = logging.getLogger(__name__)


def _get_module(package, target):
    return __import__(f'{package}.{target}', fromlist=[f'{target}'])


class Locator(object):
    @staticmethod
    def get_service(name_or_object: [str, object], metadata: dict = None, **kwargs):
        metadata = metadata or {}
        package = config.get_package()

        if isinstance(name_or_object, str):
            service_module = _get_module(package, 'service')
            return getattr(service_module, name_or_object)(metadata=metadata, **kwargs)
        else:
            return name_or_object(metadata=metadata, **kwargs)

    @staticmethod
    def get_manager(name_or_object: [str, object], **kwargs):
        package = config.get_package()
        try:
            if isinstance(name_or_object, str):
                manager_module = _get_module(package, 'manager')
                return getattr(manager_module, name_or_object)(**kwargs)
            else:
                return name_or_object(**kwargs)

        except ERROR_BASE as e:
            raise e

        except Exception as e:
            raise ERROR_LOCATOR(name=name_or_object, reason=e, _meta={'type': 'manager'})

    @staticmethod
    def get_info(name_or_object: [str, object], *args, **kwargs):
        package = config.get_package()
        try:
            if isinstance(name_or_object, str):
                info_module = _get_module(package, 'info')
                return getattr(info_module, name_or_object)(*args, **kwargs)
            else:
                return name_or_object(*args, **kwargs)

        except ERROR_BASE as e:
            raise e

        except Exception as e:
            raise ERROR_LOCATOR(name=name_or_object, reason=e, _meta={'type': 'info'})

    @staticmethod
    def get_model(name_or_object: [str, object]):
        package = config.get_package()
        try:
            if isinstance(name_or_object, str):
                model_module = _get_module(package, 'model')
                model = getattr(model_module, name_or_object)
                model.init()
                return model
            else:
                name_or_object.init()
                return name_or_object

        except ERROR_BASE as e:
            raise e

        except Exception as e:
            raise ERROR_LOCATOR(name=f'{name_or_object} Model', reason=e, _meta={'type': 'model'})

    @staticmethod
    def get_connector(name_or_object: [str, object], **kwargs):
        package = config.get_package()

        try:
            if isinstance(name_or_object, str):
                connector_conf = config.get_connector(name_or_object)
                backend = connector_conf.get('backend')

                if backend:
                    connector_module, name_or_object = backend.rsplit('.', 1)
                    connector_module = __import__(connector_module, fromlist=[name_or_object])
                else:
                    connector_module = _get_module(package, 'connector')

                _connector = getattr(connector_module, name_or_object)(config=connector_conf, **kwargs)
                return _connector
            else:
                connector_conf = config.get_connector(name_or_object.__name__)
                return name_or_object(config=connector_conf, **kwargs)

        except ERROR_BASE as e:
            raise e

        except Exception as e:
            raise ERROR_LOCATOR(name=name_or_object, reason=e, _meta={'type': 'connector'})



import logging
import inspect
from spaceone.core import config
from spaceone.core.error import *

__init__ = ['is_set', 'get', 'set', 'increment', 'decrement', 'keys', 'ttl', 'delete', 'delete_pattern',
            'flush', 'cacheable']

_CACHE_CONNECTIONS = {}
_LOGGER = logging.getLogger(__name__)


def _create_connection(backend):
    global_conf = config.get_global()
    if backend not in global_conf.get('CACHES', {}):
        raise ERROR_CACHE_CONFIGURATION(backend=backend)

    cache_conf = global_conf['CACHES'][backend].copy()
    backend = cache_conf.pop('backend', None)

    if backend is None:
        raise ERROR_CACHE_CONFIGURATION(backend=backend)

    module_name, class_name = backend.rsplit('.', 1)
    cache_module = __import__(module_name, fromlist=[class_name])
    return getattr(cache_module, class_name)(backend, cache_conf)


def connection(func):
    def wrapper(*args, backend='default', **kwargs):
        if backend not in _CACHE_CONNECTIONS:
            _CACHE_CONNECTIONS[backend] = _create_connection(backend)

        return func(_CACHE_CONNECTIONS[backend], *args, **kwargs)

    return wrapper


def _change_args_to_dict(func, args):
    args_dict = {}
    args_length = len(args)
    func_args = inspect.getfullargspec(func).args

    defaults = inspect.getfullargspec(func).defaults
    if defaults:
        first_default_index = len(func_args) - len(defaults)

    i = 0
    for arg_key in func_args:
        if i < args_length:
            args_dict[arg_key] = args[i]

        else:
            if defaults:
                args_dict[arg_key] = defaults[i - first_default_index]

        i += 1

    return args_dict


def _make_cache_key(key_format, args_dict):
    key_data = {}
    for key, value in args_dict.items():
        if isinstance(value, list) or isinstance(value, tuple):
            sorted_values = sorted(value)
            values_str = ','.join(sorted_values)
            key_data[key] = values_str
        else:
            key_data[key] = value

    try:
        return key_format.format(**key_data)

    except Exception:
        raise ERROR_CACHE_KEY_FORMAT(key=key_format)


def cacheable(key=None, value=None, expire=None, action='cache', backend='default'):
    def wrapper(func):
        def wrapped_func(*args, **kwargs):
            if is_set(backend):
                args_dict = _change_args_to_dict(func, args)
                args_dict.update(kwargs)
                cache_key = _make_cache_key(key, args_dict)
                if action in ['cache']:
                    data = get(cache_key, backend=backend)
                    if data is not None:
                        return data

            result = func(*args, **kwargs)

            if is_set(backend):
                if action in ['put', 'cache']:
                    cache_value = result
                    if value:
                        if isinstance(value, dict):
                            cache_value = result.get(value)

                        else:
                            if hasattr(result, value):
                                cache_value = getattr(result, value)
                            else:
                                raise ERROR_CACHEABLE_VALUE_TYPE()

                    set(cache_key, cache_value, expire=expire, backend=backend)

                if action in ['delete']:
                    delete(cache_key, backend=backend)

            return result

        return wrapped_func

    return wrapper


def is_set(backend='default'):
    global_conf = config.get_global()
    if global_conf.get('CACHES', {}).get(backend, {}) != {}:
        return True
    else:
        return False


@connection
def get(cache_cls, key):
    return cache_cls.get(key)


@connection
def set(cache_cls, key, value, expire=None):
    return cache_cls.set(key, value, expire=expire)


@connection
def increment(cache_cls, key, amount=1):
    return cache_cls.increment(key, amount)


@connection
def decrement(cache_cls, key, amount=1):
    return cache_cls.decrement(key, amount)


@connection
def keys(cache_cls, pattern):
    return cache_cls.keys(pattern)


@connection
def ttl(cache_cls, key):
    return cache_cls.ttl(key)


@connection
def delete(cache_cls, *keys):
    return cache_cls.delete(*keys)


@connection
def delete_pattern(cache_cls, pattern):
    return cache_cls.delete_pattern(pattern)


@connection
def flush(cache_cls, is_async=False):
    return cache_cls.flush(is_async)


class BaseCache(object):

    def get(self, key, **kwargs):
        """
        Args:
            key (str)
            **kwargs (dict)
        Returns:
            cache_value (any)
        """
        raise NotImplementedError('cache.get not implemented!')

    def set(self, key, value, **kwargs):
        """
        Args:
            key (str)
            value (str)
            **kwargs (dict)
                - expire (int: seconds)

        Returns:
            True | False
        """
        raise NotImplementedError('cache.set not implemented!')

    def increment(self, key, amount):
        """
        Args:
            key (str)
            amount (int)

        Returns:
            True | False
        """
        raise NotImplementedError('cache.incr not implemented!')

    def decrement(self, key, amount):
        """
        Args:
            key (str)
            amount (int)

        Returns:
            True | False
        """
        raise NotImplementedError('cache.decr not implemented!')

    def keys(self, pattern='*'):
        """
        Args:
            pattern (str)

        Returns:
            keys (list)
        """
        raise NotImplementedError('cache.keys not implemented!')

    def ttl(self, key):
        """
        Args:
            key (str)

        Returns:
            expire_time (int: seconds)
        """
        raise NotImplementedError('cache.ttl not implemented!')

    def delete(self, *keys):
        """
        Args:
            keys (list)

        Returns:
            None
        """
        raise NotImplementedError('cache.delete not implemented!')

    def delete_pattern(self, pattern):
        """
        Args:
            pattern (str)

        Returns:
            None
        """
        raise NotImplementedError('cache.delete_pattern not implemented!')

    def flush(self, is_async=False):
        """
        Args:
            is_async (bool)

        Returns:
            None
        """
        raise NotImplementedError('cache.flush not implemented!')



from spaceone.core.base import CoreObject


class BaseConnector(CoreObject):

    def __init__(self, config: dict = None, **kwargs):
        super().__init__()

        self.config = config or {}

        for key, value in kwargs.items():
            setattr(self, key, value)



class BaseModel(object):

    @classmethod
    def connect(cls):
        """ 
        Args:
        Returns:
            None
        """
        raise NotImplementedError('model.connect not implemented!')

    @classmethod
    def create(cls, data):
        """ 
        Args:
            data (dict)
        Returns:
            model_vo (object)
        """
        raise NotImplementedError('model.create not implemented!')

    def update(self, data):
        """ 
        Args:
            data (dict)
        Returns:
            model_vo (object)
        """
        raise NotImplementedError('model.update not implemented!')

    def delete(self):
        """
        Args:
        Returns:
            None
        """
        raise NotImplementedError('model.delete not implemented!')

    def terminate(self):
        """ 
        Args:
        Returns:
            None
        """
        raise NotImplementedError('model.terminate not implemented!')

    def increment(self, key, amount=1):
        """
        Args:
            key (str)
            amount (int)
        Returns:
            model_vo (object)
        """
        raise NotImplementedError('model.increment not implemented!')

    def decrement(self, key, amount=1):
        """
        Args:
            key (str)
            amount (int)
        Returns:
            model_vo (object)
        """
        raise NotImplementedError('model.decrement not implemented!')

    @classmethod
    def get(cls, **conditions):
        """ 
        Args:
            **conditions (kwargs)
            - key (str): value (any)
        Returns:
            model_vo (object)
        """
        raise NotImplementedError('model.get not implemented!')

    @classmethod
    def filter(cls, **conditions):
        """ 
        Args:
            **conditions (kwargs)
            - key (str): value (any)
        Returns:
            model_vos (list)
        """
        raise NotImplementedError('model.filter not implemented!')

    def to_dict(self):
        """ 
        Args:
        Returns:
            model_data (dict)
        """
        raise NotImplementedError('model.to_dict not implemented!')

    @classmethod
    def query(cls, **query):
        """
        Args:
            **query (kwargs)
                - filter (list)
                [
                    {
                        'key' : 'field (str)',
                        'value' : 'value (any)',
                        'operator' : 'lt | lte | gt | gte | eq | not | exists | contain |
                        not_contain | in | not_in | not_contain_in | match | regex | regex_in |
                        datetime_lt | datetime_lte | datetime_gt | datetime_gte |
                        timediff_lt | timediff_lte | timediff_gt | timediff_gte'
                    },
                    ...
                ]
                - filter_or (list)
                [
                    {
                        'key' : 'field (str)',
                        'value' : 'value (any)',
                        'operator' : 'lt | lte | gt | gte | eq | not | exists | contain |
                        not_contain | in | not_in | not_contain_in | match | regex | regex_in |
                        datetime_lt | datetime_lte | datetime_gt | datetime_gte |
                        timediff_lt | timediff_lte | timediff_gt | timediff_gte'
                    },
                    ...
                ]
                - sort (dict)
                {
                  'key' : 'field (str)',
                  'desc' : True | False
                }
                - page (dict)
                {
                    'start': 'start_row (int)',
                    'limit' : 'row_limit (int)'
                }
                - distinct (str): 'field'
                - only (list): ['field1', 'field2', '...']
                - exclude(list): ['field1', 'field2', '...']
                - minimal (bool)
                - count_only (bool)

        Returns:
            model_vos (list)
            total_count (int)
        """
        raise NotImplementedError('model.query not implemented!')

    @classmethod
    def stat(cls, **query):
        """
        Args:
            **query (kwargs)
                - filter (list)
                [
                    {
                        'key' : 'field (str)',
                        'value' : 'value (any)',
                        'operator' : 'lt | lte | gt | gte | eq | not | exists | contain | not_contain |
                        in | not_in | contain_in | not_contain_in | match | regex | regex_in |
                        datetime_lt | datetime_lte | datetime_gt | datetime_gte |
                        timediff_lt | timediff_lte | timediff_gt | timediff_gte'
                    },
                    ...
                ]
                - filter_or(list)
                [
                    {
                        'key' : 'field (str)',
                        'value' : 'value (any)',
                        'operator' : 'lt | lte | gt | gte | eq | not | exists | contain | not_contain |
                        in | not_in | contain_in | not_contain_in | match | regex | regex_in |
                        datetime_lt | datetime_lte | datetime_gt | datetime_gte |
                        timediff_lt | timediff_lte | timediff_gt | timediff_gte'
                    },
                    ...
                ]
                - aggregate (dict)
                {
                    'unwind': [
                        {
                            'path': 'key path (str)'
                        }
                    ],
                    'group': {
                        'keys': [
                            {
                                'key': 'field (str)',
                                'name': 'alias name (str)'
                            },
                            ...
                        ],
                        'fields': [
                            {
                                'key': 'field (str)',
                                'name': 'alias name (str)',
                                'operator': 'count | sum | avg | max | min | size | add_to_set | merge_objects'
                            },
                            ...
                        ]
                    }
                    'count': {
                        'name': 'alias name (str)'
                    }
                }
                - sort(dict)
                {
                  'name' : 'field (str)',
                  'desc' : True | False
                }
                - page(dict)
                {
                    'start': 'start_row (int)',
                    'limit' : 'row_limit (int)'
                }

        Returns:
            values (list)
        """
        raise NotImplementedError('model.stat not implemented!')


    @classmethod
    def analyze(cls, **query):
        raise NotImplementedError('model.analyze not implemented!')



import re
import logging
import types
import grpc
import copy
import ssl
import socket
from google.protobuf import descriptor_pb2
from grpc_reflection.v1alpha import reflection_pb2
from grpc_reflection.v1alpha import reflection_pb2_grpc
from spaceone.core import utils
from spaceone.core.error import *
from spaceone.core.pygrpc.message_type import get_well_known_types

_MAX_RETRIES = 2
_GRPC_CHANNEL = {}
_LOGGER = logging.getLogger(__name__)


def _list_to_kwargs(func):
    def wrapped_func(list_args):
        messages = []
        for args in list_args:
            messages.append(func(**args))

        return messages

    return wrapped_func


def _dict_to_kwargs(func):
    def wrapped_func(args):
        return func(**args)

    return wrapped_func


class _ClientInterceptor(
    grpc.UnaryUnaryClientInterceptor, grpc.UnaryStreamClientInterceptor,
    grpc.StreamUnaryClientInterceptor, grpc.StreamStreamClientInterceptor):

    def __init__(self, options, channel_key):
        self._channel_key = channel_key
        self.metadata = options.get('metadata', {})
        self._MESSAGE_TYPE_MAP = {}
        self._FIELD_TYPE_MAP = {}

    def add_message_type(self, package, module_name, service_name, method_name, input_type):
        message_name = input_type.split('.')[-1:][0]
        method_key = f'/{package}.{service_name}/{method_name}'
        pb2_module = __import__(f'{package}.{module_name}_pb2', fromlist=[f'{module_name}_pb2'])

        if method_key not in self._MESSAGE_TYPE_MAP:
            self._MESSAGE_TYPE_MAP[method_key] = {}

        well_known_types = get_well_known_types()

        if input_type in well_known_types:
            self._MESSAGE_TYPE_MAP[method_key] = {
                'message': well_known_types[input_type],
                'message_type': 'well_known_message'
            }

        elif input_type.startswith(f'.{package}'):
            if hasattr(pb2_module, message_name):
                self._MESSAGE_TYPE_MAP[method_key] = {
                    'message': getattr(pb2_module, message_name),
                    'message_name': message_name
                }

                if message_name not in self._FIELD_TYPE_MAP:
                    self._FIELD_TYPE_MAP[message_name] = {
                        'well_known_type': {},
                        'message_type': {}
                    }

    def add_wellknown_type(self, message_type_name, field_name, change_method):
        if message_type_name not in self._FIELD_TYPE_MAP:
            self._FIELD_TYPE_MAP[message_type_name] = {
                'well_known_type': {},
                'message_type': {}
            }

        self._FIELD_TYPE_MAP[message_type_name]['well_known_type'][field_name] = change_method

    def add_field_type(self, message_type_name, field_name, change_method):
        if message_type_name not in self._FIELD_TYPE_MAP:
            self._FIELD_TYPE_MAP[message_type_name] = {
                'well_known_type': {},
                'message_type': {}
            }

        self._FIELD_TYPE_MAP[message_type_name]['message_type'][field_name] = change_method

    def _change_request(self, message_info, request):
        message = message_info['message']
        message_name = message_info.get('message_name')

        if message_info.get('message_type') == 'well_known_message':
            return message(request)

        elif message_name:
            for key in self._FIELD_TYPE_MAP[message_name]['well_known_type']:
                change_method = self._FIELD_TYPE_MAP[message_name]['well_known_type'][key]
                request = utils.change_dict_value(request, key, change_method, change_type='func')

            for key, value in request.items():
                if key in self._FIELD_TYPE_MAP[message_name]['message_type']:
                    change_method = self._FIELD_TYPE_MAP[message_name]['message_type'][key]
                    request[key] = change_method(value)

            change_message = message(**request)

            return change_message

        else:
            return request

    def _make_message(self, request, method_key):
        if isinstance(request, dict):
            return self._change_request(self._MESSAGE_TYPE_MAP[method_key], copy.deepcopy(request))

        else:
            return request

    def _generate_message(self, request_iterator, method_key):
        for request in request_iterator:
            yield self._make_message(request, method_key)

    def _check_message(self, client_call_details, request_or_iterator, is_stream):
        if client_call_details.method in self._MESSAGE_TYPE_MAP:
            if is_stream:
                if not isinstance(request_or_iterator, types.GeneratorType):
                    raise Exception("Stream method must be specified as a generator type.")

                return self._generate_message(request_or_iterator, client_call_details.method)

            else:
                return self._make_message(request_or_iterator, client_call_details.method)

        return request_or_iterator

    def _check_error(self, response):
        if isinstance(response, Exception):
            details = response.details()
            status_code = response.code().name
            if details.startswith('ERROR_'):
                details_split = details.split(':', 1)
                if len(details_split) == 2:
                    error_code, error_message = details_split
                else:
                    error_code = details_split[0]
                    error_message = details

                if status_code == 'PERMISSION_DENIED':
                    raise ERROR_PERMISSION_DENIED()
                elif status_code == 'UNAUTHENTICATED':
                    raise ERROR_AUTHENTICATE_FAILURE(message=error_message)
                else:
                    e = ERROR_INTERNAL_API(message=error_message)
                    e.error_code = error_code
                    e.status_code = status_code
                    raise e

            else:
                error_message = response.details()
                if status_code == 'PERMISSION_DENIED':
                    raise ERROR_PERMISSION_DENIED()
                elif status_code == 'PERMISSION_DENIED':
                    raise ERROR_AUTHENTICATE_FAILURE(message=error_message)
                elif status_code == 'UNAVAILABLE':
                    e = ERROR_GRPC_CONNECTION(channel=self._channel_key, message=error_message)
                    e.meta['channel'] = self._channel_key
                    raise e
                else:
                    e = ERROR_INTERNAL_API(message=error_message)
                    e.status_code = status_code
                    raise e

        return response

    def _generate_response(self, response_iterator):
        try:
            for response in response_iterator:
                yield self._check_error(response)
        except Exception as e:
            self._check_error(e)

    def _retry_call(self, continuation, client_call_details, request_or_iterator, is_stream):
        retries = 0

        while True:
            try:
                response_or_iterator = continuation(client_call_details, request_or_iterator)

                if is_stream:
                    response_or_iterator = self._generate_response(response_or_iterator)
                else:
                    self._check_error(response_or_iterator)

                return response_or_iterator

            except Exception as e:
                if e.error_code == 'ERROR_GRPC_CONNECTION':
                    if retries >= _MAX_RETRIES:
                        channel = e.meta.get('channel')
                        if channel in _GRPC_CHANNEL:
                            _LOGGER.error(f'Disconnect gRPC Endpoint. (channel = {channel})')
                            del _GRPC_CHANNEL[channel]
                        raise e
                    else:
                        _LOGGER.debug(f'Retry gRPC Call: reason = {e.message}, retry = {retries + 1}')
                else:
                    raise e

            retries += 1

    def _intercept_call(self, continuation, client_call_details,
                        request_or_iterator, is_request_stream, is_response_stream):
        new_request_or_iterator = self._check_message(
            client_call_details, request_or_iterator, is_request_stream)

        return self._retry_call(continuation, client_call_details,
                                new_request_or_iterator, is_response_stream)

    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request, False, False)

    def intercept_unary_stream(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request, False, True)

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        return self._intercept_call(continuation, client_call_details, request_iterator, True, False)

    def intercept_stream_stream(self, continuation, client_call_details, request_iterator):
        return self._intercept_call(continuation, client_call_details, request_iterator, True, True)


class _GRPCClient:

    def __init__(self, channel, options, channel_key):
        self.api_resources = {}
        self._well_known_types = get_well_known_types()
        self._related_message_types = {}
        self._client_interceptor = _ClientInterceptor(options, channel_key)
        self._channel = channel
        self._intercept_channel = grpc.intercept_channel(channel, self._client_interceptor)
        self._get_server_reflection_info()

    @staticmethod
    def _parse_proto_name(proto_name, package):
        return re.findall(r'%s/(.*?).proto' % package, proto_name)[0]

    def _create_grpc_stub(self, package, module_name, service_name):
        grpc_pb2_module = __import__(f'{package}.{module_name}_pb2_grpc', fromlist=[f'{module_name}_pb2_grpc'])

        setattr(self, service_name,
                getattr(grpc_pb2_module, f'{service_name}Stub')(self._intercept_channel))

    def _change_message_field_type(self, module_name, message_type_name, field, is_repeated=False):
        pb2_module = self._related_message_types[module_name]['module']
        message_types = self._related_message_types[module_name]['message_types']

        if field.type_name in message_types:
            field_message_name = field.type_name.split('.')[-1:][0]

            if is_repeated:
                change_method = _list_to_kwargs(getattr(pb2_module, field_message_name))
            else:
                change_method = _dict_to_kwargs(getattr(pb2_module, field_message_name))

            self._client_interceptor.add_field_type(
                message_type_name,
                field.name,
                change_method)

    def _change_wellknown_field_type(self, message_type_name, field, field_name):
        change_method = self._well_known_types[field.type_name]
        self._client_interceptor.add_wellknown_type(
            message_type_name,
            field_name,
            change_method)

    def _find_well_known_type_in_field(self, module_name, message_type_name,
                                       field_type_name, parent_field_name, parent_field_label):
        if field_type_name in self._related_message_types[module_name]['message_types']:
            for field in self._related_message_types[module_name]['message_types'][field_type_name].field:
                if field.type == 11:
                    if parent_field_label == 3:  # field.label == LABEL_REPEATED
                        field_key = f'{parent_field_name}.[].{field.name}'
                    else:
                        field_key = f'{parent_field_name}.{field.name}'

                    if field.type_name in self._well_known_types:
                        self._change_wellknown_field_type(message_type_name, field, field_key)

                    else:
                        if field_type_name != field.type_name:
                            self._find_well_known_type_in_field(module_name,
                                                                message_type_name,
                                                                field.type_name,
                                                                field_key,
                                                                field.label)

    def _preload_field_type(self, message_types, module_name):
        for message_type in message_types:
            for field in message_type.field:
                if field.type == 11:  # field.type_name == TYPE_MESSAGE
                    if field.type_name in self._well_known_types:
                        if field.label == 3:  # field.label == LABEL_REPEATED
                            self._change_wellknown_field_type(message_type.name, field, f'{field.name}.[]')
                        else:
                            self._change_wellknown_field_type(message_type.name, field, field.name)

                    else:
                        self._find_well_known_type_in_field(module_name,
                                                            message_type.name,
                                                            field.type_name,
                                                            field.name,
                                                            field.label)

                        if field.label == 3:  # field.label == LABEL_REPEATED
                            self._change_message_field_type(module_name, message_type.name, field, True)
                        else:
                            self._change_message_field_type(module_name, message_type.name, field)

    def _preload_related_message_type(self, file_descriptor_proto, module_name):
        package = file_descriptor_proto.package
        if module_name not in self._related_message_types:
            self._related_message_types[module_name] = {
                'module': __import__(f'{package}.{module_name}_pb2', fromlist=[f'{module_name}_pb2']),
                'message_types': {}
            }

        for message_type in file_descriptor_proto.message_type:
            message_type_name = f'.{package}.{message_type.name}'
            self._related_message_types[module_name]['message_types'][message_type_name] = message_type

    def _preload_message_type(self, package, module_name, service_name, file_descriptor_proto):
        self.api_resources[service_name] = []
        methods = file_descriptor_proto.service[0].method
        for method in methods:
            self._client_interceptor.add_message_type(
                package,
                module_name,
                service_name,
                method.name,
                method.input_type)

            self.api_resources[service_name].append(method.name)

    def _get_file_descriptor(self, reflection_stub, service):
        # Get File Descriptor
        message = reflection_pb2.ServerReflectionRequest(file_containing_symbol=service.name)
        responses = reflection_stub.ServerReflectionInfo(iter((message,)))
        for response in responses:
            file_descriptor_proto = descriptor_pb2.FileDescriptorProto.FromString(
                response.file_descriptor_response.file_descriptor_proto[0])

            # Parse File Descriptor
            package = file_descriptor_proto.package
            service_name = file_descriptor_proto.service[0].name
            module_name = self._parse_proto_name(file_descriptor_proto.name, package)

            self._create_grpc_stub(package, module_name, service_name)
            self._preload_message_type(
                package,
                module_name,
                service_name,
                file_descriptor_proto)

            self._preload_related_message_type(file_descriptor_proto, module_name)
            self._preload_field_type(file_descriptor_proto.message_type, module_name)

    def _get_server_reflection_info(self):
        reflection_stub = reflection_pb2_grpc.ServerReflectionStub(self._channel)

        # List Services
        message = reflection_pb2.ServerReflectionRequest(list_services='')
        responses = reflection_stub.ServerReflectionInfo(iter((message,)))

        for response in responses:
            for service in response.list_services_response.service:
                self._get_file_descriptor(reflection_stub, service)


def _check_server_port(host, port, endpoint):
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(3)
        result = sock.connect_ex((host, port))
        sock.settimeout(None)

    except Exception as e:
        raise ERROR_GRPC_CONNECTION(channel=endpoint, message='failed to connect to all addresses')

    if result != 0:
        raise ERROR_GRPC_CONNECTION(channel=endpoint, message='failed to connect to all addresses')


def _parse_endpoint(endpoint):
    try:
        host, port = endpoint.split(':')
        return host, int(port)
    except Exception:
        raise ERROR_GRPC_CONNECTION(channel=endpoint, message='Endpoint format is invalid. (format = <host>:<port>)')


def _create_secure_channel(endpoint, options):
    # host, port = _parse_endpoint(endpoint)
    # _check_server_port(host, port, endpoint)

    try:
        # cert = ssl.get_server_certificate((host, port))
        # creds = grpc.ssl_channel_credentials(str.encode(cert))
        creds = grpc.ssl_channel_credentials()
    except Exception as e:
        raise ERROR_GRPC_TLS_HANDSHAKE(reason=e)

    return grpc.secure_channel(endpoint, creds, options=options)


def _create_insecure_channel(endpoint, options):
    # host, port = _parse_endpoint(endpoint)
    # _check_server_port(host, port, endpoint)

    return grpc.insecure_channel(endpoint, options=options)


def client(endpoint=None, ssl_enabled=False, max_message_length=None, **client_opts):
    if endpoint is None:
        raise Exception("Client's endpoint is undefined.")

    if endpoint not in _GRPC_CHANNEL:
        options = []

        if max_message_length:
            options.append(('grpc.max_send_message_length', max_message_length))
            options.append(('grpc.max_receive_message_length', max_message_length))

        if ssl_enabled:
            channel = _create_secure_channel(endpoint, options)
        else:
            channel = _create_insecure_channel(endpoint, options)

        try:
            grpc.channel_ready_future(channel).result(timeout=3)
        except Exception as e:
            raise ERROR_GRPC_CONNECTION(channel=endpoint, message='Channel is not ready.')

        try:
            _GRPC_CHANNEL[endpoint] = _GRPCClient(channel, client_opts, endpoint)
        except Exception as e:
            if hasattr(e, 'details'):
                raise ERROR_GRPC_CONNECTION(channel=endpoint, message=e.details())
            else:
                raise ERROR_GRPC_CONNECTION(channel=endpoint, message=str(e))

    return _GRPC_CHANNEL[endpoint]


def get_grpc_method(uri_info):
    try:
        conn = client(endpoint=uri_info['endpoint'], ssl_enabled=uri_info['ssl_enabled'])
        return getattr(getattr(conn, uri_info['service']), uri_info['method'])

    except ERROR_BASE as e:
        raise e
    except Exception as e:
        raise ERROR_GRPC_CONFIGURATION(endpoint=uri_info.get('endpoint'),
                                       service=uri_info.get('service'),
                                       method=uri_info.get('method'))



# -*- coding: utf-8 -*-

import json
import redis
from redis import SSLConnection

from spaceone.core.error import *
from spaceone.core.cache import BaseCache


class RedisCache(BaseCache):

    def __init__(self, backend, cache_conf):
        try:
            if cache_conf.get('ssl', False):
                del cache_conf['ssl']
                pool = redis.ConnectionPool(connection_class=SSLConnection, **cache_conf)
            else:
                pool = redis.ConnectionPool(**cache_conf)
            self.conn = self._get_connection(pool)
            self.conn.ping()
        except redis.exceptions.TimeoutError:
            raise ERROR_CACHE_TIMEOUT(config=cache_conf)
        except Exception:
            raise ERROR_CACHE_CONFIGURATION(backend=backend)

    def _get_connection(self, pool):
        return redis.Redis(connection_pool=pool)

    def get(self, key, **kwargs):
        try:
            cache_value = self.conn.get(key)
            if cache_value:
                return json.loads(cache_value)
            else:
                return cache_value

        except Exception as e:
            raise ERROR_CACHE_DECODE(reason=e)

    def set(self, key, value, expire=None, **kwargs):
        try:
            if value is None:
                value = {}

            cache_value = json.dumps(value)
            return self.conn.set(key, cache_value, ex=expire)
        except Exception as e:
            raise ERROR_UNKNOWN(message=e)

    def increment(self, key, amount=1):
        try:
            return self.conn.incr(key, amount)
        except Exception as e:
            raise ERROR_UNKNOWN(message=e)

    def decrement(self, key, amount=1):
        try:
            return self.conn.decr(key, amount)
        except Exception as e:
            raise ERROR_UNKNOWN(message=e)

    def keys(self, pattern='*'):
        return self.conn.keys(pattern)

    def ttl(self, key):
        return self.conn.ttl(key)

    def delete(self, *keys):
        self.conn.delete(*keys)

    def delete_pattern(self, pattern):
        for key in self.keys(pattern):
            self.conn.delete(key)

    def flush(self, is_async=False):
        self.conn.flushdb(is_async)



import logging

from spaceone.core.cache import cacheable
from spaceone.core.handler import BaseAuthenticationHandler
from spaceone.core.connector.space_connector import SpaceConnector
from spaceone.core.error import *

_LOGGER = logging.getLogger(__name__)

_EXCLUDE_METHODS = [
    'identity.APIKey.get',
    'identity.Authorization.verify'
]


class AuthenticationAPIKeyHandler(BaseAuthenticationHandler):

    def __init__(self, config):
        super().__init__(config)
        self._initialize()

    def _initialize(self):
        if 'uri' not in self.config:
            _LOGGER.error(f'[_initialize] uri config is undefined.')
            raise ERROR_HANDLER_CONFIGURATION(handler='AuthenticationGRPCHandler')

        self.client: SpaceConnector = self.locator.get_connector(SpaceConnector, endpoint=self.config['uri'])

    def verify(self, params=None):
        request_method = f'{self.transaction.service}.{self.transaction.resource}.{self.transaction.verb}'
        if request_method not in _EXCLUDE_METHODS:
            token_type = self.transaction.get_meta('token_type')
            api_key_id = self.transaction.get_meta('api_key_id')
            api_key_ver = self.transaction.get_meta('api_key_ver')
            domain_id = self.transaction.get_meta('domain_id')

            if token_type == 'API_KEY':
                if api_key_id and domain_id:
                    self._check_api_key(api_key_id, domain_id)
                else:
                    raise ERROR_AUTHENTICATE_FAILURE(message='This API Key is no longer supported.')

    @cacheable(key='api-key:{domain_id}:{api_key_id}', backend='local')
    def _check_api_key(self, api_key_id, domain_id):
        _LOGGER.debug(f'[_check_api_key] check api key state ({api_key_id})')

        try:
            response = self.client.dispatch(
                'APIKey.get',
                {
                    'api_key_id': api_key_id,
                    'domain_id': domain_id
                }
            )

            if response['state'] != 'ENABLED':
                raise ERROR_AUTHENTICATE_FAILURE(message='The state of the API key has been disabled.')

            return True

        except ERROR_AUTHENTICATE_FAILURE as e:
            raise e
        except ERROR_GRPC_CONNECTION as e:
            raise e
        except Exception as e:
            _LOGGER.error(f'[_check_api_key] API Call Error: {e}')
            raise ERROR_AUTHENTICATE_FAILURE(message='API key is invalid.')



# Database Settings
DATABASES = {
    'default': {
        # MongoDB Example
        # 'host': '<host>',
        # 'port': 27017,
        # 'db': '<db>',
        # 'username': '<user>',
        # 'password': '<password>'
    },
    # 'local': {
    #     'backend': 'spaceone.core.cache.local_cache.LocalCache',
    #     'max_size': 128,
    #     'ttl': 86400
    # }
}

# Cache Settings
CACHES = {
    'default': {
        # Redis Example
        # 'backend': 'spaceone.core.cache.redis_cache.RedisCache',
        # 'host': '<host>',
        # 'port': 6379,
        # 'db': 0
    }
}

# Handler Configuration
HANDLERS = {
    'authentication': [
        # Default Authentication Handler
        # {
        #     'backend': 'spaceone.core.handler.authentication_handler.AuthenticationGRPCHandler',
        #     'uri': 'grpc://identity:50051/v1/Domain/get_public_key'
        # }
    ],
    'authorization': [
        # Default Authorization Handler
        # {
        #     'backend': 'spaceone.core.handler.authorization_handler.AuthorizationGRPCHandler',
        #     'uri': 'grpc://identity:50051/v1/Authorization/verify'
        # }
    ],
    'mutation': [],
    'event': []
}

# Connector Settings
CONNECTORS = {
}

# Log Settings
LOG = {
}



__ALL__ = ["SPACEONE_TASK_SCHEMA"]

SPACEONE_TASK_SCHEMA = {
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "version": {"type": "string", "enum": ["v1"]},
        "executionEngine": {"type": "string"},
        "stages": {"type": "array", "items": {"$ref": "#/definitions/task"}}
    },
    "definitions": {
        "task": {
            "type": "object",
            "properties": {
                "locator": {"type": "string",
                            "enum": ["SERVICE", "MANAGER"]},
                "name": {"type": "string"},
                "metadata": {"type": "object"},
                "method": {"type": "string"},
                "params": {"type": "object"}
            },
            "required": ["locator", "name", "metadata", "method", "params"]
        }
    },
    "required": ["stages"]
}


"""
instance = {
    "name": "Publish",
    "version": "v1",
    "executionEngine": "BasePipelineWorker",
    "stages": [
        {
            "locator": "SERVICE",
            "name": "SupervisorService",
            "metadata": [["token", "1234"]],
            "method": "publish",
            "params": {"params": {"id": "aaaa"}}
        }
    ]
}

instance = {
    "name": 1
}

instance = {
    "name": "Publish",
    "version": "v1",
    "executionEngine": "BasePipelineWorker",
    "stages": [
        {
            "locator": "MANAGER",
            "name": "SupervisorService",
            "params": {"name": "aaaa", "metadata": 111}
        }
    ]
}


print(validate(instance, schema=SPACEONE_TASK_SCHEMA))
"""



import logging
import fnmatch

from spaceone.core.handler import BaseAuthorizationHandler
from spaceone.core.connector.space_connector import SpaceConnector
from spaceone.core.error import ERROR_HANDLER_CONFIGURATION, ERROR_PERMISSION_DENIED

_LOGGER = logging.getLogger(__name__)


class AuthorizationGRPCHandler(BaseAuthorizationHandler):

    def __init__(self, config):
        super().__init__(config)
        self.client = None
        self._initialize()

    def _initialize(self):
        if 'uri' not in self.config:
            _LOGGER.error(f'[_initialize] uri config is undefined.')
            raise ERROR_HANDLER_CONFIGURATION(handler='AuthenticationGRPCHandler')

        self.client: SpaceConnector = self.locator.get_connector(SpaceConnector, endpoint=self.config['uri'])

    def verify(self, params=None):
        self._check_permissions()

        user_type = self.transaction.get_meta('authorization.user_type')
        scope = self.transaction.get_meta('authorization.scope', 'DOMAIN')

        if user_type == 'DOMAIN_OWNER':
            self._verify_domain_owner(params)
        else:
            self._verify_auth(params, scope)

    def _check_permissions(self):
        permissions = self.transaction.get_meta('authorization.permissions')
        if isinstance(permissions, list):
            for permission in permissions:
                request_api = f'{self.transaction.service}.{self.transaction.resource}.{self.transaction.verb}'
                if fnmatch.fnmatch(request_api, permission):
                    return True

            raise ERROR_PERMISSION_DENIED()

        return True

    def _verify_domain_owner(self, params):
        # Pass all methods
        self.transaction.set_meta('authorization.role_type', 'DOMAIN')

    def _verify_auth(self, params, scope):
        project_id_key = self.transaction.get_meta('authorization.project_id', 'project_id')
        project_group_id_key = self.transaction.get_meta('authorization.project_group_id', 'project_group_id')
        user_id_key = self.transaction.get_meta('authorization.user_id', 'user_id')
        require_project_group_id = self.transaction.get_meta('authorization.require_project_group_id', False)
        require_project_id = self.transaction.get_meta('authorization.require_project_id', False)
        require_user_id = self.transaction.get_meta('authorization.require_user_id', False)
        require_domain_id = self.transaction.get_meta('authorization.require_domain_id', False)

        try:
            response = self.client.dispatch(
                'Authorization.verify',
                {
                    'service': self.transaction.service,
                    'resource': self.transaction.resource,
                    'verb': self.transaction.verb,
                    'scope': scope,
                    'domain_id': params.get('domain_id'),
                    'project_id': params.get(project_id_key),
                    'project_group_id': params.get(project_group_id_key),
                    'user_id': params.get(user_id_key),
                    'require_project_id': require_project_id,
                    'require_project_group_id': require_project_group_id,
                    'require_user_id': require_user_id,
                    'require_domain_id': require_domain_id
                }
            )

            self.transaction.set_meta('authorization.role_type', response['role_type'])
            self.transaction.set_meta('authorization.projects', response.get('projects', []))
            self.transaction.set_meta('authorization.project_groups', response.get('project_groups', []))
        except Exception as e:
            _LOGGER.error(f'[_verify_auth] Authorization.verify request failed: {e}')
            raise ERROR_PERMISSION_DENIED()



from functools import reduce
from mongoengine.queryset.visitor import Q

from spaceone.core.error import *
from spaceone.core import utils

__all__ = ['FILTER_OPERATORS']


def _default_resolver(key, value, operator, is_multiple):
    if is_multiple:
        return reduce(lambda x, y: x | y,
                      map(lambda i: Q(**{f'{key}__{operator}': i}), value))
    else:
        return Q(**{f'{key}__{operator}': value})


def _eq_resolver(key, value, operator, is_multiple):
    return Q(**{key: value})


def _in_resolver(key, value, operator, is_multiple):
    return Q(**{f'{key}__in': value})


def _exists_resolver(key, value, operator, is_multiple):
    if not isinstance(value, bool):
        raise ERROR_OPERATOR_BOOLEAN_TYPE(operator=operator,
                                          condition={'key': key, 'value': value, 'operator': operator})

    return Q(**{f'{key}__exists': value})


def _not_in_resolver(key, value, operator, is_multiple):
    return Q(**{f'{key}__nin': value})


def _match_resolver(key, value, operator, is_multiple):
    if not isinstance(value, dict):
        raise ERROR_OPERATOR_DICT_VALUE_TYPE(operator=operator,
                                             condition={'key': key, 'value': value, 'operator': operator})

    return Q(**{f'{key}__match': value})


def _regex_resolver(key, value, operator, is_multiple):
    if is_multiple:
        return reduce(lambda x, y: x | y,
                      map(lambda i: Q(**{'__raw__': {key: {'$regex': i, '$options': 'i'}}}), value))
    else:
        return Q(**{'__raw__': {key: {'$regex': value, '$options': 'i'}}})


def _datetime_resolver(key, value, operator, is_multiple):
    try:
        dt = utils.iso8601_to_datetime(value)
    except Exception as e:
        raise ERROR_DB_QUERY(reason=f'The value of datetime_{operator} operator is required ISO 8601 format.')
    return Q(**{f'{key}__{operator}': dt})


def _timediff_resolver(key, value, operator, is_multiple):
    try:
        dt = utils.parse_timediff_query(value)
    except Exception as e:
        raise ERROR_DB_QUERY(reason=f'The value of timediff_{operator} operator is invalid. (value = {value})')
    return Q(**{f'{key}__{operator}': dt})


FILTER_OPERATORS = {
    # model operator : (resolver, mongoengine operator, is_multiple)
    'lt': (_default_resolver, 'lt', False),
    'lte': (_default_resolver, 'lte', False),
    'gt': (_default_resolver, 'gt', False),
    'gte': (_default_resolver, 'gte', False),
    'eq': (_eq_resolver, None, False),
    'not': (_default_resolver, 'ne', False),
    'exists': (_exists_resolver, 'exists', False),
    'contain': (_default_resolver, 'icontains', False),
    'not_contain': (_default_resolver, 'not__icontains', False),
    'in': (_in_resolver, None, True),
    'not_in': (_not_in_resolver, None, True),
    'contain_in': (_default_resolver, 'icontains', True),
    'not_contain_in': (_default_resolver, 'not__icontains', True),
    'match': (_match_resolver, 'match', False),
    'regex': (_regex_resolver, None, False),
    'regex_in': (_regex_resolver, None, True),
    'datetime_gt': (_datetime_resolver, 'gt', False),
    'datetime_lt': (_datetime_resolver, 'lt', False),
    'datetime_gte': (_datetime_resolver, 'gte', False),
    'datetime_lte': (_datetime_resolver, 'lte', False),
    'timediff_gt': (_timediff_resolver, 'gt', False),
    'timediff_lt': (_timediff_resolver, 'lt', False),
    'timediff_gte': (_timediff_resolver, 'gte', False),
    'timediff_lte': (_timediff_resolver, 'lte', False),
}



import logging
from datetime import datetime

import six
from spaceone.api.core.v1 import query_pb2, handler_pb2
from google.protobuf import struct_pb2
from google.protobuf.empty_pb2 import Empty
from google.protobuf.timestamp_pb2 import Timestamp

__all__ = ['change_value_type', 'change_struct_type', 'change_list_value_type', 'change_timestamp_type',
           'change_handler_authorization_response', 'change_query', 'change_stat_query', 'change_analyze_query',
           'change_time_series_analyze_query', 'get_well_known_types']

_INT_OR_FLOAT = six.integer_types + (float,)
LOGGER = logging.getLogger(__name__)


def change_value_type(value):
    if value is None:
        return {'null_value': 'NULL_VALUE'}
    elif isinstance(value, bool):
        return {'bool_value': value}
    elif isinstance(value, six.string_types):
        return {'string_value': value}
    elif isinstance(value, _INT_OR_FLOAT):
        return {'number_value': value}
    elif isinstance(value, dict):
        change_value = {'struct_value': struct_pb2.Struct()}
        change_value['struct_value'].update(value)
        return change_value
    elif isinstance(value, list):
        change_value = {'list_value': struct_pb2.ListValue()}
        change_value['list_value'].extend(value)
        return change_value
    else:
        return value


def change_struct_type(value):
    if isinstance(value, dict):
        change_value = struct_pb2.Struct()
        change_value.update(value)
        return change_value
    else:
        return value


def change_list_value_type(value):
    if isinstance(value, list):
        change_value = struct_pb2.ListValue()
        change_value.extend(value)
        return change_value
    else:
        return value


def change_timestamp_type(value):
    if isinstance(value, datetime):
        ts = Timestamp()
        ts.FromDatetime(value)
        return ts
    elif isinstance(value, str):
        ts = Timestamp()
        ts.FromJsonString(value)
        return ts
    else:
        return value


def change_empty_type(value):
    return Empty()


def _change_condition_type(condition):
    if 'value' in condition:
        condition['value'] = change_value_type(condition['value'])
    elif 'v' in condition:
        condition['v'] = change_value_type(condition['v'])

    return condition


def change_query(value):
    change_value = value.copy()
    change_value['filter'] = map(_change_condition_type, value.get('filter', []))
    change_value['filter_or'] = map(_change_condition_type, value.get('filter_or', []))

    return query_pb2.Query(**change_value)


def change_stat_query(value):
    change_value = value.copy()
    change_value['filter'] = map(_change_condition_type, value.get('filter', []))
    change_value['filter_or'] = map(_change_condition_type, value.get('filter_or', []))

    for stage in change_value.get('aggregate', []):
        if 'group' in stage:
            for field in stage['group'].get('fields', []):
                if 'conditions' in field:
                    field['conditions'] = map(_change_condition_type, field.get('conditions', []))

    return query_pb2.StatisticsQuery(**change_value)


def change_analyze_query(value):
    change_value = value.copy()
    change_value['filter'] = map(_change_condition_type, value.get('filter', []))
    change_value['filter_or'] = map(_change_condition_type, value.get('filter_or', []))

    if 'fields' in value:
        change_value['fields'] = change_struct_type(value['fields'])

    if 'select' in value:
        change_value['select'] = change_struct_type(value['select'])

    return query_pb2.AnalyzeQuery(**change_value)


def change_time_series_analyze_query(value):
    change_value = value.copy()
    change_value['filter'] = map(_change_condition_type, value.get('filter', []))
    change_value['filter_or'] = map(_change_condition_type, value.get('filter_or', []))

    if 'fields' in value:
        change_value['fields'] = change_struct_type(value['fields'])

    if 'select' in value:
        change_value['select'] = change_struct_type(value['select'])

    return query_pb2.TimeSeriesAnalyzeQuery(**change_value)


def change_handler_authentication_request(value):
    return handler_pb2.AuthenticationRequest(**value)


def change_handler_authorization_request(value):
    return handler_pb2.AuthorizationRequest(**value)


def change_handler_authorization_response(value):
    return handler_pb2.AuthorizationResponse(**value)


def get_well_known_types():
    return {
        '.google.protobuf.Value': change_value_type,
        '.google.protobuf.Struct': change_struct_type,
        '.google.protobuf.ListValue': change_list_value_type,
        '.google.protobuf.Timestamp': change_timestamp_type,
        '.google.protobuf.Empty': change_empty_type,
        '.spaceone.api.core.v1.Query': change_query,
        '.spaceone.api.core.v1.StatisticsQuery': change_stat_query,
        '.spaceone.api.core.v1.AnalyzeQuery': change_analyze_query,
        '.spaceone.api.core.v1.TimeSeriesAnalyzeQuery': change_time_series_analyze_query,
        '.spaceone.api.core.v1.AuthorizationRequest': change_handler_authorization_request,
        '.spaceone.api.core.v1.AuthenticationRequest': change_handler_authentication_request,
    }
